@@ -2,34 +2,46 @@ from __future__ import annotations
 
 import os.path
 from contextlib import suppress
+from copy import copy
 from typing import List, TypeVar, Generic, Optional, Dict, Any
 
-from pykotor.common.stream import BinaryReader
+from pykotor.resource.formats.bwm.bwm_auto import bytes_bwm
+from pykotor.resource.formats.erf import ERF, write_erf, read_erf
+from pykotor.resource.formats.rim import read_rim, write_rim
+
+from pykotor.resource.formats.vis.vis_auto import bytes_vis
+
+from pykotor.resource.formats.tpc.tpc_auto import bytes_tpc
+
+from pykotor.common.misc import CaseInsensitiveDict
+
+from pykotor.common.stream import BinaryReader, BinaryWriter
 from pykotor.extract.capsule import Capsule
 from pykotor.extract.file import ResourceIdentifier
 from pykotor.extract.installation import Installation, SearchLocation
+from pykotor.resource.formats.bwm import read_bwm
 from pykotor.resource.formats.gff import read_gff
 from pykotor.resource.formats.lyt import LYT
-from pykotor.resource.formats.lyt.lyt_auto import read_lyt
+from pykotor.resource.formats.lyt.lyt_auto import read_lyt, bytes_lyt
 from pykotor.resource.formats.mdl import MDL
-from pykotor.resource.formats.tpc import read_tpc
-from pykotor.resource.formats.vis import read_vis
-from pykotor.resource.generics.are import ARE, construct_are
-from pykotor.resource.generics.dlg import construct_dlg
-from pykotor.resource.generics.git import GIT, construct_git
-from pykotor.resource.generics.ifo import IFO, construct_ifo
-from pykotor.resource.generics.pth import construct_pth
-from pykotor.resource.generics.utc import UTC, construct_utc
-from pykotor.resource.generics.utd import UTD, construct_utd
-from pykotor.resource.generics.ute import UTE, construct_ute
-from pykotor.resource.generics.uti import UTI, construct_uti
-from pykotor.resource.generics.utm import UTM, construct_utm
-from pykotor.resource.generics.utp import UTP, construct_utp
-from pykotor.resource.generics.uts import construct_uts
-from pykotor.resource.generics.utt import UTT, construct_utt
-from pykotor.resource.generics.utw import UTW, construct_utw
+from pykotor.resource.formats.tpc import read_tpc, TPC
+from pykotor.resource.formats.vis import read_vis, VIS
+from pykotor.resource.generics.are import ARE, construct_are, read_are, bytes_are
+from pykotor.resource.generics.dlg import construct_dlg, read_dlg, bytes_dlg
+from pykotor.resource.generics.git import GIT, construct_git, read_git, bytes_git
+from pykotor.resource.generics.ifo import IFO, construct_ifo, read_ifo, bytes_ifo
+from pykotor.resource.generics.pth import construct_pth, read_pth, bytes_pth
+from pykotor.resource.generics.utc import UTC, construct_utc, read_utc, bytes_utc
+from pykotor.resource.generics.utd import UTD, construct_utd, read_utd, bytes_utd
+from pykotor.resource.generics.ute import UTE, construct_ute, read_ute, bytes_ute
+from pykotor.resource.generics.uti import UTI, construct_uti, read_uti, bytes_uti
+from pykotor.resource.generics.utm import UTM, construct_utm, read_utm, bytes_utm
+from pykotor.resource.generics.utp import UTP, construct_utp, read_utp, bytes_utp
+from pykotor.resource.generics.uts import construct_uts, read_uts, bytes_uts, UTS
+from pykotor.resource.generics.utt import UTT, construct_utt, read_utt, bytes_utt
+from pykotor.resource.generics.utw import UTW, construct_utw, read_utw, bytes_utw
 from pykotor.resource.type import ResourceType
-from pykotor.tools.model import list_textures
+from pykotor.tools.model import list_textures, list_lightmaps
 
 T = TypeVar('T')
 SEARCH_ORDER = [SearchLocation.OVERRIDE, SearchLocation.CUSTOM_MODULES, SearchLocation.CHITIN]
@@ -46,8 +58,7 @@ class Module:
         self._root = root
 
         self._capsules = [custom_capsule] if custom_capsule is not None else []
-        self._capsules.extend(
-            [Capsule(installation.module_path() + module) for module in installation.module_names() if root in module])
+        self._capsules.extend([Capsule(installation.module_path() + module) for module in installation.module_names() if root in module])
 
         for capsule in self._capsules:
             if capsule.exists("module", ResourceType.IFO):
@@ -57,7 +68,7 @@ class Module:
         else:
             raise ValueError("Unable to locate module IFO file for '{}'.".format(root))
 
-        self.resources: Dict[str, ModuleResource] = {}
+        self.resources: CaseInsensitiveDict[ModuleResource] = CaseInsensitiveDict()
         self.reload_resources()
 
     @staticmethod
@@ -81,6 +92,17 @@ class Module:
             rootb = rootb[:rootb.index("_")]
         return roota + rootb
 
+    def capsules(
+            self
+    ) -> List[Capsule]:
+        """
+        Returns a copy of the capsules used by the module.
+
+        Returns:
+            A list of linked capsules.
+        """
+        return copy(self._capsules)
+
     def reload_resources(
             self
     ):
@@ -97,9 +119,9 @@ class Module:
                 self.add_locations(resource.resname(), resource.restype(), [resource.filepath()])
 
         # Any resource linked in the GIT not present in the module files
-        original = self.git().active
+        original = self.git().active()
         look_for = []
-        for location in self.git().locations:
+        for location in self.git().locations():
             self.git().activate(location)
             git = self.git().resource()
             [look_for.append(ResourceIdentifier(creature.resref.get(), ResourceType.UTC)) for creature in git.creatures]
@@ -112,16 +134,18 @@ class Module:
              git.encounters]
             [look_for.append(ResourceIdentifier(trigger.resref.get(), ResourceType.UTT)) for trigger in git.triggers]
             [look_for.append(ResourceIdentifier(store.resref.get(), ResourceType.UTM)) for store in git.stores]
-        self.git().active = original
+        self.git().activate(original)
 
         # Models referenced in LYTs
-        original = self.layout().active
-        for location in self.layout().locations:
+        original = self.layout().active()
+        for location in self.layout().locations():
             self.layout().activate(location)
             layout = self.layout().resource()
             for room in layout.rooms:
                 look_for.append(ResourceIdentifier(room.model, ResourceType.MDL))
-        self.layout().active = original
+                look_for.append(ResourceIdentifier(room.model, ResourceType.MDX))
+                look_for.append(ResourceIdentifier(room.model, ResourceType.WOK))
+        self.layout().activate(original)
 
         search = self._installation.locations(look_for, [SearchLocation.OVERRIDE, SearchLocation.CHITIN])
         for identifier, locations in search.items():
@@ -132,8 +156,11 @@ class Module:
         textures = set()
         for model in self.models():
             with suppress(Exception):
-                for texture in list_textures(model.data()):
+                data = model.data()
+                for texture in list_textures(data):
                     textures.add(texture)
+                for lightmap in list_lightmaps(data):
+                    textures.add(lightmap)
         for texture in textures:
             look_for.append(ResourceIdentifier(texture, ResourceType.TPC))
             look_for.append(ResourceIdentifier(texture, ResourceType.TGA))
@@ -154,7 +181,10 @@ class Module:
             restype: ResourceType,
             locations: List[str]
     ):
-        filename = "{}.{}".format(resname, restype)
+        # In order to store TGA resources in the same ModuleResource as their TPC counterpart, we use the .TPC extension
+        # instead of the .TGA for the dictionary key.
+        filename_ext = str(ResourceType.TPC if restype == ResourceType.TGA else restype)
+        filename = "{}.{}".format(resname, filename_ext)
         if filename in self.resources:
             self.resources[filename].add_locations(locations)
         else:
@@ -166,39 +196,47 @@ class Module:
     ) -> Installation:
         return self._installation
 
+    def resource(
+            self,
+            resname: str,
+            restype: ResourceType
+    ) -> Optional[ModuleResource]:
+        filename = resname + "." + restype.extension
+        return self.resources[filename] if filename in self.resources else None
+
     def layout(
             self
     ) -> ModuleResource[LYT]:
         for filename, resource in self.resources.items():
-            if resource.resname.lower() == self._id and resource.restype() == ResourceType.LYT:
+            if resource.resname().lower() == self._id and resource.restype() == ResourceType.LYT:
                 return resource
 
     def vis(
             self
     ) -> ModuleResource[VIS]:
         for filename, resource in self.resources.items():
-            if resource.resname.lower() == self._id and resource.restype() == ResourceType.VIS:
+            if resource.resname().lower() == self._id and resource.restype() == ResourceType.VIS:
                 return resource
 
     def are(
             self
     ) -> ModuleResource[ARE]:
         for filename, resource in self.resources.items():
-            if resource.resname.lower() == self._id and resource.restype() == ResourceType.ARE:
+            if resource.resname().lower() == self._id and resource.restype() == ResourceType.ARE:
                 return resource
 
     def git(
             self
     ) -> ModuleResource[GIT]:
         for filename, resource in self.resources.items():
-            if resource.resname.lower() == self._id and resource.restype() == ResourceType.GIT:
+            if resource.resname().lower() == self._id and resource.restype() == ResourceType.GIT:
                 return resource
 
     def info(
             self
     ) -> ModuleResource[IFO]:
         for filename, resource in self.resources.items():
-            if resource.resname.lower() == "module" and resource.restype() == ResourceType.IFO:
+            if resource.resname().lower() == "module" and resource.restype() == ResourceType.IFO:
                 return resource
 
     def creature(
@@ -206,7 +244,7 @@ class Module:
             resname: str
     ) -> Optional[ModuleResource[UTC]]:
         for resource in self.resources.values():
-            if resname == resource.resname and resource.restype() == ResourceType.UTC:
+            if resname == resource.resname() and resource.restype() == ResourceType.UTC:
                 return resource
         return None
 
@@ -224,7 +262,7 @@ class Module:
             resname: str
     ) -> Optional[ModuleResource[UTP]]:
         for resource in self.resources.values():
-            if resname == resource.resname and resource.restype() == ResourceType.UTP:
+            if resname == resource.resname() and resource.restype() == ResourceType.UTP:
                 return resource
         return None
 
@@ -242,7 +280,7 @@ class Module:
             resname: str
     ) -> Optional[ModuleResource[UTD]]:
         for resource in self.resources.values():
-            if resname == resource.resname and resource.restype() == ResourceType.UTD:
+            if resname == resource.resname() and resource.restype() == ResourceType.UTD:
                 return resource
         return None
 
@@ -260,7 +298,7 @@ class Module:
             resname: str
     ) -> Optional[ModuleResource[UTI]]:
         for resource in self.resources.values():
-            if resname == resource.resname and resource.restype() == ResourceType.UTI:
+            if resname == resource.resname() and resource.restype() == ResourceType.UTI:
                 return resource
         return None
 
@@ -278,7 +316,7 @@ class Module:
             resname: str
     ) -> Optional[ModuleResource[UTE]]:
         for resource in self.resources.values():
-            if resname == resource.resname and resource.restype() == ResourceType.UTE:
+            if resname == resource.resname() and resource.restype() == ResourceType.UTE:
                 return resource
         return None
 
@@ -296,7 +334,7 @@ class Module:
             resname: str
     ) -> Optional[ModuleResource[UTM]]:
         for resource in self.resources.values():
-            if resname == resource.resname and resource.restype() == ResourceType.UTM:
+            if resname == resource.resname() and resource.restype() == ResourceType.UTM:
                 return resource
         return None
 
@@ -314,7 +352,7 @@ class Module:
             resname: str
     ) -> Optional[ModuleResource[UTT]]:
         for resource in self.resources.values():
-            if resname == resource.resname and resource.restype() == ResourceType.UTT:
+            if resname == resource.resname() and resource.restype() == ResourceType.UTT:
                 return resource
         return None
 
@@ -332,7 +370,7 @@ class Module:
             resname: str
     ) -> Optional[ModuleResource[UTW]]:
         for resource in self.resources.values():
-            if resname == resource.resname and resource.restype() == ResourceType.UTW:
+            if resname == resource.resname() and resource.restype() == ResourceType.UTW:
                 return resource
         return None
 
@@ -350,7 +388,7 @@ class Module:
             resname: str
     ) -> Optional[ModuleResource[MDL]]:
         for resource in self.resources.values():
-            if resname == resource.resname and resource.restype() == ResourceType.MDL:
+            if resname == resource.resname() and resource.restype() == ResourceType.MDL:
                 return resource
         return None
 
@@ -359,7 +397,7 @@ class Module:
             resname: str
     ) -> Optional[ModuleResource]:
         for resource in self.resources.values():
-            if resname == resource.resname and resource.restype() == ResourceType.MDX:
+            if resname == resource.resname() and resource.restype() == ResourceType.MDX:
                 return resource
         return None
 
@@ -372,6 +410,42 @@ class Module:
                 models.append(resource)
         return models
 
+    def texture(
+            self,
+            resname: str
+    ) -> Optional[ModuleResource[TPC]]:
+        for resource in self.resources.values():
+            if resname.lower() == resource.resname().lower() and resource.restype() in [ResourceType.TPC, ResourceType.TGA]:
+                return resource
+        return None
+
+    def textures(
+            self
+    ) -> List[ModuleResource[MDL]]:
+        textures = []
+        for resource in self.resources.values():
+            if resource.restype() in [ResourceType.TPC, ResourceType.TGA]:
+                textures.append(resource)
+        return textures
+
+    def sound(
+            self,
+            resname: str
+    ) -> Optional[ModuleResource[UTS]]:
+        for resource in self.resources.values():
+            if resname == resource.resname() and resource.restype() == ResourceType.UTS:
+                return resource
+        return None
+
+    def sounds(
+            self
+    ) -> List[ModuleResource[UTS]]:
+        sounds = []
+        for resource in self.resources.values():
+            if resource.restype() == ResourceType.UTS:
+                sounds.append(resource)
+        return sounds
+
 
 class ModuleResource(Generic[T]):
     def __init__(
@@ -409,6 +483,29 @@ class ModuleResource(Generic[T]):
         """
         return self._restype
 
+    def localized_name(
+            self
+    ) -> Optional[str]:
+        res = self.resource()
+        if res is None:
+            return None
+        elif isinstance(res, UTC):
+            return self._installation.string(res.first_name) + " " + self._installation.string(res.last_name)
+        elif isinstance(res, UTP):
+            return self._installation.string(res.name)
+        elif isinstance(res, UTD):
+            return self._installation.string(res.name)
+        elif isinstance(res, UTW):
+            return self._installation.string(res.name)
+        elif isinstance(res, UTT):
+            return self._installation.string(res.name)
+        elif isinstance(res, UTE):
+            return self._installation.string(res.name)
+        elif isinstance(res, UTM):
+            return self._installation.string(res.name)
+        else:
+            return None
+
     def data(
             self
     ) -> bytes:
@@ -422,7 +519,7 @@ class ModuleResource(Generic[T]):
             The bytes data of the active file.
         """
 
-        if self.active is None:
+        if self._active is None:
             raise ValueError("No file is currently active for resource '{}.{}'.".format(self.resname, self._restype.extension))
         elif self._active.endswith(".erf") or self._active.endswith(".mod") or self._active.endswith(".rim"):
             capsule = Capsule(self._active)
@@ -434,7 +531,7 @@ class ModuleResource(Generic[T]):
 
     def resource(
             self
-    ) -> T:
+    ) -> Optional[T]:
         """
         Returns the cached resource object. If no object has been cached, then it will load the object.
 
@@ -444,37 +541,39 @@ class ModuleResource(Generic[T]):
 
         if self._resource is None:
             conversions = {
-                ResourceType.UTC: (lambda data: construct_utc(read_gff(data))),
-                ResourceType.UTP: (lambda data: construct_utp(read_gff(data))),
-                ResourceType.UTD: (lambda data: construct_utd(read_gff(data))),
-                ResourceType.UTI: (lambda data: construct_uti(read_gff(data))),
-                ResourceType.UTM: (lambda data: construct_utm(read_gff(data))),
-                ResourceType.UTE: (lambda data: construct_ute(read_gff(data))),
-                ResourceType.UTT: (lambda data: construct_utt(read_gff(data))),
-                ResourceType.UTW: (lambda data: construct_utw(read_gff(data))),
-                ResourceType.UTS: (lambda data: construct_uts(read_gff(data))),
-                ResourceType.DLG: (lambda data: construct_dlg(read_gff(data))),
-                ResourceType.PTH: (lambda data: construct_pth(read_gff(data))),
+                ResourceType.UTC: (lambda data: read_utc(data)),
+                ResourceType.UTP: (lambda data: read_utp(data)),
+                ResourceType.UTD: (lambda data: read_utd(data)),
+                ResourceType.UTI: (lambda data: read_uti(data)),
+                ResourceType.UTM: (lambda data: read_utm(data)),
+                ResourceType.UTE: (lambda data: read_ute(data)),
+                ResourceType.UTT: (lambda data: read_utt(data)),
+                ResourceType.UTW: (lambda data: read_utw(data)),
+                ResourceType.UTS: (lambda data: read_uts(data)),
+                ResourceType.DLG: (lambda data: read_dlg(data)),
+                ResourceType.PTH: (lambda data: read_pth(data)),
                 ResourceType.NCS: (lambda data: data),
                 ResourceType.TPC: (lambda data: read_tpc(data)),
+                ResourceType.TGA: (lambda data: read_tpc(data)),
                 ResourceType.LYT: (lambda data: read_lyt(data)),
                 ResourceType.VIS: (lambda data: read_vis(data)),
-                ResourceType.IFO: (lambda data: construct_ifo(read_gff(data))),
-                ResourceType.ARE: (lambda data: construct_are(read_gff(data))),
-                ResourceType.GIT: (lambda data: construct_git(read_gff(data)))
+                ResourceType.IFO: (lambda data: read_ifo(data)),
+                ResourceType.ARE: (lambda data: read_are(data)),
+                ResourceType.GIT: (lambda data: read_git(data)),
+                ResourceType.WOK: (lambda data: read_bwm(data))
             }
 
             if self._active is None:
                 self._resource = None
             elif self._active.endswith(".erf") or self._active.endswith(".mod") or self._active.endswith(".rim"):
                 data = Capsule(self._active).resource(self._resname, self._restype)
-                self._resource = conversions[self.restype](data)
+                self._resource = conversions[self._restype](data)
             elif self._active.endswith("bif"):
                 data = self._installation.resource(self._resname, self._restype, [SearchLocation.CHITIN]).data
-                self._resource = conversions[self.restype](data)
+                self._resource = conversions[self._restype](data)
             else:
                 data = BinaryReader.load_file(self._active)
-                self._resource = conversions[self.restype](data)
+                self._resource = conversions[self._restype](data)
 
         return self._resource
 
@@ -490,6 +589,13 @@ class ModuleResource(Generic[T]):
             filepaths: A list of filepaths pointing to a location for the resource.
         """
         self._locations.extend([filepath for filepath in filepaths if filepath not in self._locations])
+        if self._active is None and self._locations:
+            self.activate(self._locations[0])
+
+    def locations(
+            self,
+    ) -> List[str]:
+        return self._locations
 
     def activate(
             self,
@@ -506,11 +612,28 @@ class ModuleResource(Generic[T]):
         """
         self._resource = None
         if filepath is None:
-            self._active = self.locations[0] if len(self.locations) > 0 else None
-        elif filepath in self.locations:
+            self._active = self._locations[0] if len(self._locations) > 0 else None
+        elif filepath in self._locations:
             self._active = filepath
         else:
-            raise ValueError("The filepath '{}' is not being tracked as a location for the resource.".format(active))
+            raise ValueError("The filepath '{}' is not being tracked as a location for the resource.".format(self._active))
+
+    def unload(
+            self
+    ) -> None:
+        """
+        Clears the cached resource object from memory.
+        """
+        self._resource = None
+
+    def reload(
+            self
+    ) -> None:
+        """
+        Reloads the resource object from the active location.
+        """
+        self._resource = None
+        self.resource()
 
     def active(
             self
@@ -522,3 +645,44 @@ class ModuleResource(Generic[T]):
             Filepath to the active resource.
         """
         return self._active
+
+    def save(
+            self
+    ) -> None:
+        conversions = {
+            ResourceType.UTC: (lambda res: bytes_utc(res)),
+            ResourceType.UTP: (lambda res: bytes_utp(res)),
+            ResourceType.UTD: (lambda res: bytes_utd(res)),
+            ResourceType.UTI: (lambda res: bytes_uti(res)),
+            ResourceType.UTM: (lambda res: bytes_utm(res)),
+            ResourceType.UTE: (lambda res: bytes_ute(res)),
+            ResourceType.UTT: (lambda res: bytes_utt(res)),
+            ResourceType.UTW: (lambda res: bytes_utw(res)),
+            ResourceType.UTS: (lambda res: bytes_uts(res)),
+            ResourceType.DLG: (lambda res: bytes_dlg(res)),
+            ResourceType.PTH: (lambda res: bytes_pth(res)),
+            ResourceType.NCS: (lambda res: res),
+            ResourceType.TPC: (lambda res: bytes_tpc(res)),
+            ResourceType.TGA: (lambda res: bytes_tpc(res)),
+            ResourceType.LYT: (lambda res: bytes_lyt(res)),
+            ResourceType.VIS: (lambda res: bytes_vis(res)),
+            ResourceType.IFO: (lambda res: bytes_ifo(res)),
+            ResourceType.ARE: (lambda res: bytes_are(res)),
+            ResourceType.GIT: (lambda res: bytes_git(res)),
+            ResourceType.WOK: (lambda res: bytes_bwm(res))
+        }
+
+        if self._active is None:
+            raise ValueError("No active file selected for resource '{}.{}'".format(self._resname, self._restype.extension))
+        elif self._active.endswith(".erf") or self._active.endswith(".mod"):
+            erf = read_erf(self._active)
+            erf.set(self._resname, self._restype, conversions[self._restype](self.resource()))
+            write_erf(erf, self._active)
+        elif self._active.endswith(".rim"):
+            rim = read_rim(self._active)
+            rim.set(self._resname, self._restype, conversions[self._restype](self.resource()))
+            write_rim(rim, self._active)
+        elif self._active.endswith("bif"):
+            raise ValueError("Cannot save file to BIF.")
+        else:
+            BinaryWriter.dump(self._active, conversions[self._restype](self.resource()))
